#ifndef AlienCommand_h
#define AlienCommand_h

// $Id: AlienCommand.h,v 1.8 2012/07/23 08:41:25 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "Counter.h"
#include "Debug.h"
#include "QOrderedSet.h"

#include <QtCore/QString>
#include <QtCore/QStringList>
#include <QtCore/QHash>
#include <QtCore/QList>

//! handles top commands
class AlienCommand: public Counter
{

    public:

    //! enumeration for command type
    enum Type
    {

        NONE,
        COPY,
        CHILDREN_LIST,
        DEFAULT_USER,
        JOB_FILES,
        KILL,
        KILL_MASTER,
        MASTER_LIST,
        RESUBMIT,
        RESUBMIT_MASTER,
        TRACE,
        TRACE_JDL,
        TRACE_ALL,
        USER_LIST,
        CLEAR_QUEUE
    };

    //! list of commands
    typedef QList<AlienCommand> List;

    //! constructor
    AlienCommand( const Type = NONE );

    //! less than operator
    virtual bool operator < ( const AlienCommand& other ) const
    {
        if( priority() != other.priority() ) return priority() < other.priority();
        else return id() < other.id();
    }

    //! equal to operator
    bool operator == (const AlienCommand& other ) const
    { return priority() == other.priority() && id() == other.id(); }

    //!@name accessors
    //@{

    //! command name
    QString name( void ) const
    { return name( type() ); }

    //! map Type to command name
    static QString name( const Type& type );

    //! id
    int id( void ) const
    { return id_; }

    //! priority
    int priority( void ) const
    { return priority_; }

    //! type
    Type type( void ) const
    { return type_; }

    // true if arguments are passed
    bool hasArguments( void ) const
    { return !arguments_.isEmpty(); }

    //! arguments
    const QStringList& arguments( void ) const
    { return arguments_; }

    //! user
    QString user( void ) const;

    //! master job id
    int parentJobId( void ) const;

    //! job id
    int jobId( void ) const;

    //! path
    QString path( void ) const;

    //@}

    //!@name modifiers
    //@{

    //! priority
    void setPriority( unsigned int value )
    { priority_ = value; }

    //! add argument
    AlienCommand& addArgument( const QString& argument )
    {
        arguments_.push_back( argument );
        return *this;
    }

    //! push arguments
    AlienCommand& operator << (const QString& argument )
    { return addArgument( argument ); }

    //! clear arguments
    void clearArguments( void )
    { arguments_.clear(); }

    //@}

    private:

    //! static id counter
    static unsigned int& _counter( void );

    //! map types to command name
    typedef QHash<Type,QString> NameMap;

    //! name map
    static NameMap& _names( void );

    //! unique id
    int id_;

    //! priority
    int priority_;

    //! type
    Type type_;

    //! arguments
    QStringList arguments_;

    //! initialize command names
    static void _initializeNames( void );

};

//! used to sort commands based on priority and type
class SortedAlienCommand: public AlienCommand
{

    public:

    //! constructor
    SortedAlienCommand( void )
    {}

    //! constructor
    SortedAlienCommand( const AlienCommand& other ):
        AlienCommand( other )
        {}

    virtual bool operator < (const SortedAlienCommand& other ) const
    {

        if( priority() != other.priority() ) return priority() < other.priority();
        if( type() != other.type() ) return type() < other.type();

        // check arguments
        for( int index = 0; index < std::min( arguments().size(), other.arguments().size() ); ++index )
        { if( arguments()[index] != other.arguments()[index] ) return arguments()[index] < other.arguments()[index]; }

            // check arguments size
        return arguments().size() < other.arguments().size();

    };

    typedef QOrderedSet<SortedAlienCommand> Set;

};

#endif
