// $Id: AlienInterface.cpp,v 1.15 2012/07/25 12:57:57 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

/*!
\file    AlienInterface.cc
\brief   Alien commands interface
\author  Hugo Pereira
\version $Revision: 1.15 $
\date    $Date: 2012/07/25 12:57:57 $
*/


#include "AlienInterface.h"

#include "Application.h"
#include "CustomProcess.h"
#include "Command.h"
#include "Debug.h"
#include "File.h"
#include "FileRecordProperties.h"
#include "IOString.h"
#include "JobSet.h"
#include "Singleton.h"
#include "TextFormat.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QtCore/QTextStream>
#include <QtCore/QRegExp>
#include <algorithm>

//_______________________________________________
AlienInterface::AlienInterface( QObject *parent ):
    QObject( parent ),
    Counter( "AlienInterface" ),
    paused_( false ),
    hideKilledJobs_( false ),
    process_( 0 )
{
    Debug::Throw("AlienInterface::AlienInterface.\n");
    Application& application( *static_cast<Application*>(Singleton::get().application()) );
    connect( &application, SIGNAL( configurationChanged() ), SLOT( _updateConfiguration() ) );
    _updateConfiguration();
}

//_______________________________________________
AlienInterface::~AlienInterface( void )
{ Debug::Throw("AlienInterface::~AlienInterface.\n"); }

//__________________________________________________________________________________
void AlienInterface::addCommand( const AlienCommand& command )
{

    Debug::Throw( "AlienInterface::addCommand.\n" );

    // avoid duplicates
    /* this is innefficient */
    const bool empty( commands_.empty() );

    // doe nothing if command is not effectively added (duplicate)
    if( !commands_.append( command ) ) return;

    // emit signal
    emit commandAdded( command );

    // run command if there is none pending
    if( empty ) _runCommand();

}

//__________________________________________________________________________________
void AlienInterface::addCommands( const AlienCommand::List& commands )
{

    Debug::Throw( "AlienInterface::addCommands.\n" );

    // list of added commands
    AlienCommand::List addedCommands;

    // avoid duplicates
    /* this is innefficient */
    const bool empty( commands_.empty() );
    foreach( const AlienCommand& command, commands )
    {
        if( commands_.append( command ) )
        { addedCommands << command; }

    }

    // check if commands have been effectively added
    if( !addedCommands.empty() )
    {
        emit commandsAdded( addedCommands );
        if( empty ) _runCommand();
    }

}

//__________________________________________________________________________________
void AlienInterface::abortCommands( const AlienCommand::List& commands )
{
    Debug::Throw( "AlienInterface::abortCommands.\n" );
    foreach( const AlienCommand& command, commands )
    { if( commands_.abort( command ) ) emit commandCompleted( command ); }
}

//____________________________________________________________
void AlienInterface::abortCurrentCommand( void )
{
    Debug::Throw( "AlienInterface::abortCurrentCommand.\n" );
    if( !process_ || process_->state() == QProcess::NotRunning )
    { return; }

    process_->kill();

}

//____________________________________________________________
void AlienInterface::clearQueue( const QString& user, const Job::IdSet& jobs )
{

    Debug::Throw() << "Alieninterface::clearQueue - user: " << user << " jobs: " << jobs.size() << endl;

    // list of commands to be aborted.
    AlienCommand::List abortedCommands;

    // get all commands from queue
    SortedAlienCommand::Set commands( commands_.commands() );
    foreach( const SortedAlienCommand& command, commands )
    {
        // check user
        if( command.user() != user ) continue;

        // check job and parent id, for removal
        if(
            ( command.jobId() >= 0 && !jobs.contains( command.jobId() ) ) ||
            ( command.parentJobId() >= 0 && !jobs.contains( command.parentJobId() ) ) )
        { abortedCommands << command; }

    }

    if( !abortedCommands.isEmpty() )
    { abortCommands( abortedCommands ); }

}

//____________________________________________________________
void AlienInterface::_error( QProcess::ProcessError error )
{

    assert( !commands_.empty() );
    AlienCommand command( commands_.front() );
    Debug::Throw() << "AlienInterface::_error - command: " << command.name() << " (" << command.id() << ")" << endl;

    QString buffer;
    QTextStream( &buffer, QIODevice::WriteOnly ) << "failed to " << command.name();
    emit messageAvailable( buffer );
    emit logAvailable( buffer+"\n\n", FORMAT::ITALIC|FORMAT::BOLD );
    return;
}

//______________________________________________________________
void AlienInterface::_commandCompleted( int return_code, QProcess::ExitStatus status )
{

    assert( !commands_.empty() );

    AlienCommand command( commands_.front() );
    Debug::Throw() << "AlienInterface::_commandCompleted - command: " << command.name() << " (" << command.id() << ")" << endl;
    if( status != QProcess::NormalExit || _process().error() == QProcess::FailedToStart || _commandFailed() )
    {

        QString buffer;
        QTextStream( &buffer, QIODevice::WriteOnly ) << "failed to " << command.name();
        emit messageAvailable( buffer );
        emit logAvailable( buffer + "\n\n", FORMAT::ITALIC|FORMAT::BOLD );

        emit commandCompleted( command );
        commands_.popFront();

        _runCommand();
        return;

    }

    // update status
    if( !commands_.hasPendingCommands() )
    {
        QString buffer;
        QTextStream( &buffer ) << "Last update: " << TimeStamp::now().toString( TimeStamp::TIME_LONG );
        emit messageAvailable( buffer );
    }

    // add (cosmetics) empty line to log.
    emit logAvailable( "\n" );

    // run action depending on current type
    switch( command.type() )
    {

        case AlienCommand::DEFAULT_USER:
        {

            _processDefaultUser();
            break;
        }

        case AlienCommand::USER_LIST:
        {

            _processUserList();
            break;
        }

        case AlienCommand::MASTER_LIST:
        {
            _processMasterList( command.user() );
            break;
        }

        case AlienCommand::CHILDREN_LIST:
        {
            Job job;
            job.assign( Job::USER, command.user() )
                .assign( Job::ID, command.jobId() );
            _processChildrenList( job );
            break;
        }

        case AlienCommand::JOB_FILES:
        {
            Job job;
            job.setParentId( command.parentJobId() )
                .assign( Job::USER, command.user() )
                .assign( Job::ID, command.jobId() );
            _processJobFiles( job, command.arguments()[3] );
             break;
        }

        case AlienCommand::COPY:
        {

            // run external application, if provided
            if( command.arguments().size() >= 3 )
            { ( Command( command.arguments()[2] ) << command.arguments()[1] ).run(); }

            break;
        }

        case AlienCommand::TRACE:
        case AlienCommand::TRACE_ALL:
        {
            Job job;
            job.setParentId( command.parentJobId() )
                .assign( Job::USER, command.user() )
                .assign( Job::ID, command.jobId() );
            _processJobInformation( job, command.type() == AlienCommand::TRACE_ALL );
            break;
        }

        case AlienCommand::TRACE_JDL:
        {
            Job job;
            job.setParentId( command.parentJobId() )
                .assign( Job::USER, command.user() )
                .assign( Job::ID, command.jobId() );
            _processJdl( job );
            break;
        }

        default: break;

    }

    // emit signal, remove from pile and run next command if any
    emit commandCompleted( command );
    commands_.popFront();
    _runCommand();

    return;

}

//______________________________________________________________
void AlienInterface::_stateChanged( QProcess::ProcessState state )
{}

//________________________________________________________________________
void AlienInterface::_processDefaultUser( void )
{

    Debug::Throw( "AlienInterface::_processDefaultUser.\n" );
    QString defaultUser;

    // for some reason default user is written to stderr
    QTextStream stream( &stderrBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    while( !( line = stream.readLine() ).isNull() )
    {
        if( _isError( line ) ) continue;
        defaultUser = line.trimmed();
    }

    emit messageAvailable( "Default user available" );
    emit defaultUserAvailable( defaultUser );

}

//________________________________________________________________________
void AlienInterface::_processUserList( void )
{

    Debug::Throw( "AlienInterface::_processUserList.\n" );
    UserSet users;

    // parse output
    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;

    while( !( line = stream.readLine() ).isNull() )
    { if( !( line = line.trimmed() ).isEmpty() ) users.insert( line ); }

    emit messageAvailable( "User list available" );
    emit userListAvailable( users );

}

//________________________________________________________________________
void AlienInterface::_processMasterList( const QString& user )
{

    Debug::Throw( "AlienInterface::_processMasterList.\n" );

    // list of found jobs
    JobSet jobs;
    JobSet killedJobs;

    // parse output
    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    while( !( line = stream.readLine() ).isNull() )
    {

        if( line.isEmpty() ) continue;

        // create job from line
        Job job(line.trimmed());

        // check validity
        if( !job.isValid() ) continue;

        // check user
        if( job.get( Job::USER ) != user ) continue;

        // store job in relevant job sets
        if( job.isKilled() )
        {

            killedJobs.insert( job );
            if( !hideKilledJobs_ ) jobs.insert( job );

        } else  jobs.insert( job );

    }

    // store killed jobs
    killedJobs_[user] = killedJobs;

    emit messageAvailable( "Master job list available" );
    emit masterListAvailable( user, jobs );

}

//________________________________________________________________________
void AlienInterface::_processChildrenList( const Job& master )
{

    Debug::Throw( "AlienInterface::_processJobList.\n" );

    // list of found jobs
    JobSet jobs;
    JobSet killedJobs;

    // get user
    const QString user( master.get( Job::USER ) );

    // parse output
    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    while( !( line = stream.readLine() ).isNull() )
    {

        if( line.isEmpty() ) continue;

        // create job from line
        Job job(line.trimmed());

        // check validity
        if( !job.isValid() ) continue;

        // check user
        if( job.get( Job::USER ) != user ) continue;

        // assign parent Id
        if( !job.isChild() ) continue;

        // assign parentId
        job.setParentId( master.id() );

        // store job in relevant job sets
        if( job.isKilled() )
        {

            killedJobs.insert( job );
            if( !hideKilledJobs_ ) jobs.insert( job );

        } else  jobs.insert( job );

    }

    // store killed jobs
    killedJobs_[user] = killedJobs;

    emit childrenListAvailable( master, jobs );

}

//________________________________________________________________________
void AlienInterface::_processJobFiles( const Job& job, const QString& path )
{
    Debug::Throw( "AlienInterface::_processJobFiles" );

    // list of files
    FileRecord::List files;

    // regular expression
    QRegExp fileRegExp(
        "(\\S+)\\s+"                                      // permissions
        "(\\w+)\\s+(\\w+)\\s+"                            // user and group
        "(\\d+)\\s+"                                      // size
        "((?:\\w+)\\s+(?:\\d+)\\s+(?:\\d+):(?:\\d+))\\s+" // date
        "(\\S+)" );

    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    while( !( line = stream.readLine() ).isNull() )
    {

        // parse line
        if( fileRegExp.indexIn(line) < 0 ) continue;

        FileRecord file;
        file.setFile( File( fileRegExp.cap(6) ).addPath( path ) );
        file.addProperty( FileRecordProperties::PERMISSIONS, fileRegExp.cap(1) );
        file.addProperty( FileRecordProperties::USER, fileRegExp.cap(2) );
        file.addProperty( FileRecordProperties::GROUP, fileRegExp.cap(3) );
        file.addProperty( FileRecordProperties::SIZE, fileRegExp.cap(4) );
        file.addProperty( FileRecordProperties::CREATION, fileRegExp.cap(5) );

        files.push_back( file );

    }

    emit jobFilesAvailable( job, path, files );

}

//________________________________________________________________________
void AlienInterface::_processJobInformation( Job job, bool emitTrace )
{

    Debug::Throw( "AlienInterface::_processJobInformation.\n" );

    // regular expression to match data
    // date string is: Tue Nov  1 15:49:06 2011
    QRegExp timeRegExp( "(\\w+)\\s+(\\w+)\\s+(\\d+)\\s+(\\d+):(\\d+):(\\d+)\\s+(\\d+)\\s+" );

    // parent ID regExp  [Master Job is 135064557]
    QRegExp parentRegExp( "\\[Master Job is (\\d+)\\]" );

    // child ID regExp
    QRegExp childRegExp( "Subjob submitted:\\s+(\\d+)" );
    QRegExp resubmissionRegExp( "resubmiting subjob\\s+(\\d+)\\s+\\((\\d+)\\)" );
    QRegExp newJobRegExp( "New job \\(resubmission from (\\d+)\\)" );

    // line type enumeration
    enum LineType
    {
        NONE = 0,
        STARTING = 1<<0,
        RUNNING = 1<<1,
        COMPLETED = 1<<2
    };

    // id of resubmitted job
    int resubmitted( -1 );

    // loop over lines, and look for trace information
    bool first( true );

    // parse output
    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    while( !( line = stream.readLine() ).isNull() )
    {

        if( line.indexOf( "[state ]" ) >= 0 )
        {

            // parent ID
            if( parentRegExp.indexIn(line) >= 0 )
            { job.setParentId( parentRegExp.cap(1).toInt() ); }

            // resubmission
            if( newJobRegExp.indexIn(line) >= 0 )
            { resubmitted = newJobRegExp.cap(1).toInt(); }

            // job time stamps
            // decide line type
            unsigned int lineType( NONE );
            if( first ) lineType |= STARTING;

            if(
                line.indexOf( "to SPLIT" ) >= 0 ||
                line.indexOf( "to ASSIGNED" ) >= 0 ||
                line.indexOf( "to RUNNING" ) >= 0 ||
                line.indexOf( "to STARTED" ) >= 0 ) lineType |= RUNNING;

            else if( line.indexOf( "to ERROR" ) >= 0 || line.indexOf( "to DONE" ) >= 0 ) lineType |= COMPLETED;

            // skip if no valid type set
            if( lineType == NONE ) continue;

            // parse date
            if( timeRegExp.indexIn(line) >= 0 )
            {

                QString date;
                QTextStream( &date ) << timeRegExp.cap(2) << " " << timeRegExp.cap(3) << " " << timeRegExp.cap(4) << ":" << timeRegExp.cap(5);

                // assign to job
                first = false;
                if( lineType&STARTING ) job.assign( Job::SUB_TIME, date );
                if( lineType&RUNNING ) job.assign( Job::RUN_TIME, date );
                if( lineType&COMPLETED ) job.assign( Job::END_TIME, date );

            }

        }

    }

    // process job resubmission
    if( resubmitted >= 0 )
    {
        // find previous job from killed jobs list, and reassign parent
        Job sibling = _findKilledJob(  job.get( Job::USER ), resubmitted );
        if( sibling.isValid() && sibling.hasParent() )
        { job.setParentId( sibling.parentId() ); }
    }

    if( emitTrace ) emit jobInformationAvailable( job, stdoutBuffer_.text() );
    else emit jobInformationAvailable( job );

    return;
}

//________________________________________________________________________
void AlienInterface::_processJdl( const Job& job )
{
    QRegExp regExp( "^\\s{8,8}" );

    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    QString buffer;
    while( !( line = stream.readLine() ).isNull() )
    {
        if( line == "    [" || line == "    ]" ) continue;
        if( line.indexOf( regExp ) < 0 ) continue;
        line.replace( regExp, "" );
        buffer+= line + "\n";
    }

    // find output directory
    QString outputDir;
    QRegExp outputDirRegExp( "OutputDir\\s+=\\s+\\\"(\\S+)\\\"" );
    if( outputDirRegExp.indexIn( buffer ) >= 0 )
    {
        outputDir = outputDirRegExp.cap(1);
        Debug::Throw() << "AlienInterface::_processJdl - output directory: " << outputDir << endl;
    }

    // find output files
    QStringList outputFiles;
    QRegExp outputFilesRegExp( "OutputFile\\s+=\\s+\\{\\s+(\\\"\\S+\\\",*\\s+)+\\}" );
    QRegExp fileRegExp( "\\\"(\\S+)\\\"" );
    if( outputFilesRegExp.indexIn( buffer ) >= 0 )
    {
        // capture strings from matched text
        const QString capturedText( outputFilesRegExp.cap(0) );

        int position = 0;
        while(( position = fileRegExp.indexIn( capturedText, position ) ) >= 0 )
        {
            outputFiles << fileRegExp.cap(1);
            position += fileRegExp.matchedLength();
        }

        Debug::Throw() << "AlienInterface::_processJdl - output files: " << outputFiles.join( " " ) << endl;

    }

    // TODO:
    // - add HTML markups to buffer;
    // - build output files and pass to JdlAvailable
    emit jdlAvailable( job, buffer );
}

//___________________________________________________________________
void AlienInterface::_readStdout(  void )
{
    _process().setReadChannel( QProcess::StandardOutput );
    IOString buffer( _process() );

    // replace dummy characters
    const QRegExp regExp( "(\\S){2,2}((\\d{2,2};\\d{2,2})|(0))m" );
    buffer.replace( regExp, "" );

    emit logAvailable( buffer );
    stdoutBuffer_.append( buffer );
}

//___________________________________________________________________
void AlienInterface::_readStderr(  void )
{
    _process().setReadChannel( QProcess::StandardError );
    IOString buffer( _process() );
    emit logAvailable( buffer, FORMAT::ITALIC );
    stderrBuffer_.append( buffer );
}

//_______________________________________
void AlienInterface::_updateConfiguration( void )
{
    Debug::Throw( "AlienInterface::_updateConfiguration.\n" );

    // get system environment
    systemEnvironment_ = QProcessEnvironment::systemEnvironment();

    // get certificates and check
    const File certificates( XmlOptions::get().raw( "CERTIFICATES_PATH" ) );
    if( certificates.exists() && certificates.isDirectory() )
    { systemEnvironment_.insert( "XrdSecSSLCADIR", certificates ); }

    // get gshell root path from gbbox binary
    const QString gbboxCommand( XmlOptions::get().raw( "GBBOX" ) );
    const QString match( "/bin/gbbox" );
    const int position( gbboxCommand.lastIndexOf( match ) );
    File gShellPath;
    if( position == gbboxCommand.length() - match.length() )
    { gShellPath = gbboxCommand.left( position ) + "/"; }

    // check
    if( gShellPath.exists() && gShellPath.isDirectory() )
    { systemEnvironment_.insert( "GSHELL_ROOT", gShellPath ); }

}

//____________________________________________________________
void AlienInterface::_resetBuffers( void )
{
    Debug::Throw( "AlienInterface::_resetBuffer.\n" );
    stdoutBuffer_.clear();
    stderrBuffer_.clear();
}

//__________________________________________________________________________________
bool AlienInterface::_runCommand( void )
{

    // do nothing if paused
    if( _paused() ) return false;

    // do nothing if no commands are in the pile
    if( commands_.empty() )
    {

        Debug::Throw( "AlienInterface::_runCommand - ready \n" );
        emit ready();
        return false;

    }

    // front command
    AlienCommand command( commands_.front() );
    Debug::Throw() << "AlienInterface::_runCommand - command: " << command.name() << " (" << command.id() << ")" << endl;

    /*
    don't run command if it is associated to a user that is not in the
    list of active users any more
    */
    if(
        (
        command.type() == AlienCommand::MASTER_LIST ||
        command.type() == AlienCommand::CHILDREN_LIST ||
        command.type() == AlienCommand::TRACE ) &&
        activeUsers_.find( command.user() ) == activeUsers_.end() )
    {
        emit commandCompleted( command );
        commands_.popFront();
        return _runCommand();
    }

    // reset buffer
    _resetBuffers();

    // take action depending on command type
    switch( command.type() )
    {

        case AlienCommand::DEFAULT_USER:
        {

            emit logAvailable( "Requesting default user\n", FORMAT::BOLD );
            emit messageAvailable( "Requesting default user" );

            QStringList arguments;
            arguments << XmlOptions::get().raw("ALIEN_WHOAMI" );

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::USER_LIST:
        {

            emit logAvailable( "Requesting user list\n", FORMAT::BOLD );
            emit messageAvailable( "Requesting user list" );

            QStringList arguments;
            arguments << XmlOptions::get().raw("ALIEN_WHO" );

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::MASTER_LIST:
        {

            // check arguments
            assert( command.hasArguments() );

            // get user and check
            const QString user( command.user() );
            if( user.isEmpty() ) break;

            // emit message
            QString buffer;
            QTextStream( &buffer ) << "Requesting job list for user " << user;
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("ALIEN_PS")
                << "-F" <<  "l" << "-b"
                << "-u" << user;

            if( command.type() == AlienCommand::MASTER_LIST )
            { arguments << "-M"; }

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::CHILDREN_LIST:
        {

            // check arguments
            const QString user( command.user() );
            const int id( command.jobId() );

            // emit message
            QString buffer;
            QTextStream( &buffer ) << "Requesting children list for job " << id;
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );

            // create command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("ALIEN_PS")
                << "-F" <<  "l" << "-b"
                << "-m" << QString().setNum( id );

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::JOB_FILES:
        {


            // get id
            const int id( command.jobId() );
            const QString path( command.path() );

            // emit message
            QString buffer;
            QTextStream( &buffer ) << "Requesting output files list for job " << id;
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw( "GBBOX" )
                << "ls" << "-n" << "-l" << path;

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::COPY:
        {

            assert( command.arguments().size() >= 2 );
            const QString remote( command.arguments()[0] );
            const QString local( command.arguments()[1] );

            // emit message
            QString buffer;
            QTextStream( &buffer ) << "Copying file " << remote;
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("ALIEN_CP")
                << QString( "alien:" ) + remote
                << QString( "file:" ) + local;

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );

            return true;
        }

        case AlienCommand::TRACE:
        case AlienCommand::TRACE_ALL:
        {

            // get id
            const int parentId( command.parentJobId() );
            const int id( command.jobId() );

            // emit message
            QString buffer;
            QTextStream( &buffer ) << "Requesting info for job " << id;
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("ALIEN_PS")
                << "-trace"
                << QString().setNum( id );

            // add output selection string based on command type
            if( command.type() == AlienCommand::TRACE )
            {

                // request all state lines
                arguments << "state";

                // all so request "submission" lines, to get list of children
                if( parentId < 0 ) arguments << "submit";

            } else if( command.type() == AlienCommand::TRACE_ALL ) arguments << "all";

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::TRACE_JDL:
        {

            // get id
            const int id( command.jobId() );

            // emit message
            QString buffer;
            QTextStream( &buffer ) << "Requesting JDL for job " << id;
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("ALIEN_PS")
                << "-jdl"
                << QString().setNum( id );

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::RESUBMIT:
        {

            // get user and check
            const int id( command.jobId() );
            if( id < 0 ) break;

            // emit message
            QString buffer;
            QTextStream( &buffer ) << "Resubmitting job " << id;
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );

            QStringList arguments;
            arguments << XmlOptions::get().raw( "ALIEN_RESUBMIT" ) << QString().setNum( id );

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;
        }

        case AlienCommand::RESUBMIT_MASTER:
        {

            // check arguments
            if( !command.hasArguments() ) break;

            // get user and check
            const int id( command.jobId() );
            if( id < 0 ) break;

            // get status from arguments, if any
            const QString status( command.arguments().size() >= 4 ? command.arguments()[3]:"ERROR_ALL" );

            // run and emit signal
            QString buffer;
            QTextStream( &buffer ) << "Resubmitting failed jobs for job " << id << " (status: " << status << ")";
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );

            QStringList arguments;
            arguments << XmlOptions::get().raw( "GBBOX" ) << "masterJob" << QString().setNum( id ) << "-status" << status << "resubmit";

            // start process
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::KILL:
        {

            // get user and check
            const int id( command.jobId() );

            // emit message
            QString buffer;
            QTextStream( &buffer ) << "Killing job " << id;
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );

            QStringList arguments;
            arguments << XmlOptions::get().raw( "ALIEN_KILL" ) << QString().setNum( id );

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;
        }

        case AlienCommand::KILL_MASTER:
        {

            // get id
            const int id( command.jobId() );
            if( id < 0 ) break;

            // emit message
            QString buffer;
            QTextStream( &buffer ) << "Killing job " << id;
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments << XmlOptions::get().raw( "GBBOX" ) << "masterJob" << QString().setNum(id);

            // add status
            if( command.arguments().size() >= 4 )
            { arguments << "-status" << command.arguments()[3]; }

            arguments << "kill";

            // run and emit signal
            _newProcess().start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::CLEAR_QUEUE:
        {

            const QString user( command.user() );
            assert( !user.isEmpty() );

            // emit message
            QString buffer;
            QTextStream( &buffer ) << "Clearing command queue for user " << user;
            emit logAvailable( buffer + "\n", FORMAT::BOLD );
            emit messageAvailable( buffer );
            break;
        }

        default:
        break;
    }

    emit commandCompleted( command );
    commands_.popFront();

    return _runCommand();

}

//____________________________________________________________
CustomProcess& AlienInterface::_newProcess( void )
{

    Debug::Throw( "AlienInterface::_newProcess.\n" );
    if( process_ ) process_->deleteLater();

    // create new process and assign customized environment
    process_ = new CustomProcess( this );
    process_->setProcessEnvironment( systemEnvironment_ );

    connect( &_process(), SIGNAL( stateChanged( QProcess::ProcessState ) ), this, SLOT( _stateChanged( QProcess::ProcessState ) ) );
    connect( &_process(), SIGNAL( error(  QProcess::ProcessError ) ), SLOT( _error( QProcess::ProcessError ) ) );
    connect( &_process(), SIGNAL( finished( int, QProcess::ExitStatus ) ), this, SLOT( _commandCompleted( int, QProcess::ExitStatus ) ) );
    connect( &_process(), SIGNAL( readyReadStandardOutput() ), SLOT( _readStdout() ) );
    connect( &_process(), SIGNAL( readyReadStandardError() ), SLOT( _readStderr() ) );
    return _process();

}

//____________________________________________________________
bool AlienInterface::_commandFailed( void ) const
{
    Debug::Throw( "AlienInterface::_commandFailed.\n" );
    if( stderrBuffer_.text().indexOf( "Unable to execute - session invalid -" ) >= 0 ) return true;
    if( stderrBuffer_.text().indexOf( "The URL provided is incorrect" ) >= 0 ) return true;
    return false;
}

//____________________________________________________________
bool AlienInterface::_isError( const QString& value ) const
{
    if( value.indexOf( "verify failed" ) >= 0 ) return true;
    else if( value.indexOf( "err -1" ) >= 0 ) return true;
    else if( value.indexOf( "err 10" ) >= 0 ) return true;
    else return false;
}

//____________________________________________________________
Job AlienInterface::_findKilledJob( const QString& user, int id ) const
{
    JobMap::const_iterator iter( killedJobs_.find( user ) );
    if( iter == killedJobs_.end() ) return Job();

    Job local;
    local.assign( Job::ID, QString().setNum( id ) );
    JobSet::const_iterator jobIter( iter.value().find( local ) );
    return jobIter == iter.value().end() ? Job():*jobIter;
}
