// $Id: JobManager.cpp,v 1.34 2012/07/23 13:52:04 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

/*!
\file    JobManager.cc
\brief   Job managment
\author  Hugo Pereira
\version $Revision: 1.34 $
\date    $Date: 2012/07/23 13:52:04 $
*/

#include "JobManager.h"

#include "AnimatedTreeView.h"
#include "Command.h"
#include "Debug.h"
#include "File.h"
#include "Job.h"
#include "InformationDialog.h"
#include "Str.h"
#include "UserSet.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QtGui/QApplication>
#include <QtGui/QHeaderView>
#include <QtGui/QScrollBar>
#include <QtCore/QTextStream>

//_______________________________________________
JobManager::JobManager( QWidget *parent ):
    QWidget( parent ),
    user_( UserSet::defaultUser() ),
    requestsLocked_( false )
{
    Debug::Throw("JobManager::JobManager.\n");

    // layout
    setLayout( new QVBoxLayout( this ) );
    layout()->setSpacing(0);
    layout()->setMargin(0);

    // list
    list_ = new TreeView( this );

    layout()->addWidget( list_ );
    list().setModel( &model_ );
    list().setSortingEnabled( true );
    list().setRootIsDecorated( true );
    list().setSelectionMode( TreeView::ContiguousSelection );
    list().setOptionName( "JOBLIST" );

    connect( &model_, SIGNAL( layoutAboutToBeChanged() ), SLOT( _storeSelectedJobs() ) );
    connect( &model_, SIGNAL( layoutAboutToBeChanged() ), SLOT( _storeExpandedJobs() ) );

    connect( &list(), SIGNAL( expanded( const QModelIndex& ) ), &list(), SLOT( resizeColumns() ) );
    connect( &list(), SIGNAL( expanded( const QModelIndex& ) ), SLOT( _requestChildrenList( const QModelIndex& ) ) );
    connect( &list(), SIGNAL( collapsed( const QModelIndex& ) ), &list(), SLOT( resizeColumns() ) );

    connect( &model_, SIGNAL( layoutChanged() ), SLOT( _restoreSelectedJobs() ) );
    connect( &model_, SIGNAL( layoutChanged() ), SLOT( _restoreExpandedJobs() ) );

}

//_______________________________________________
JobManager::~JobManager( void )
{ Debug::Throw( "JobManager::~JobManager.\n"); }

//_______________________________________
bool JobManager::setUser( QString user )
{

    Debug::Throw() << "JobManager::setUser - " << user << endl;
    if( user.isEmpty() || user == user_ ) return false;
    user_ = user;
    model_.clear();
    return true;

}

//____________________________________________________________
JobSet JobManager::updateMasterList( JobSet jobs )
{
    Debug::Throw( "JobManager::updateMasterList.\n" );

    // update timestamps from model
    _updateTimeStamps( jobs );
    _updateChildrenStatistics( jobs );

    // keep track of jobs that are missing time stamps
    const JobSet incompleteJobs( _incompleteJobs( jobs ) );

    // keep tracks of jobs to be deleted from model
    JobModel::List deletedJobs;
    Job::IdSet deletedIds;

     // retrieve current set
    foreach( const Job& currentJob, model_.children() )
    {

        // skip children
        if( currentJob.hasParent() )
        {

            if( deletedIds.contains( currentJob.parentId() ) )
            {
                deletedJobs << currentJob;
                deletedIds.insert( currentJob.id() );
            }

        } else if( !jobs.contains( currentJob ) ) {

            deletedJobs << currentJob;
            deletedIds.insert( currentJob.id() );

        }

    }

    // busy
    emit busy( jobs.size() );

    // removed jobs that are not found
    model_.remove( deletedJobs );

    // assign to model
    model_.add( jobs.toList() );

    // re-emit
    emit jobListAvailable( user(), JobSet().fromList( model_.children() ) );

    // iddle
    emit idle();

    QString buffer;
    QTextStream( &buffer ) << "Last update: " << TimeStamp::now().toString( TimeStamp::TIME_LONG );
    emit messageAvailable( buffer );

    return incompleteJobs;
}

//____________________________________________________________
JobSet JobManager::updateChildrenList( const Job& master, JobSet jobs )
{

    Debug::Throw( "JobManager::updateChildrenList.\n" );
    _updateTimeStamps( jobs, false );
    _updateChildrenStatistics( master, jobs );

    // keep track of jobs that are missing time stamps
    const JobSet incompleteJobs( _incompleteJobs( jobs ) );

    // assign to model
    model_.set( master, jobs.toList() );

    // re-emit
    emit jobListAvailable( user(), JobSet().fromList( model_.children() ), false );

    return incompleteJobs;
}

//____________________________________________________________
JobSet JobManager::updateJobList( JobSet jobs )
{
    Debug::Throw( "JobManager::updateJobList.\n" );

    // update timestamps
    _updateTimeStamps( jobs );

    // keep track of jobs that are missing time stamps
    const JobSet incompleteJobs( _incompleteJobs( jobs ) );

    // busy
    emit busy( jobs.size() );

    // check children (this updates master jobs statistics)
    jobs.checkChildren();

    // assign to model
    model_.set( jobs.toList() );

    // re-emit
    emit jobListAvailable( user(), jobs );

    // iddle
    emit idle();

    QString buffer;
    QTextStream( &buffer ) << "Last update: " << TimeStamp::now().toString( TimeStamp::TIME_LONG );
    emit messageAvailable( buffer );

    return incompleteJobs;
}

//____________________________________________________________
void JobManager::updateJob( const Job& job )
{

    Debug::Throw() << "JobManager::updateJob - job " << job.id() << endl;

    // get index from job
    QModelIndex index( model_.index( job ) );
    if( !index.isValid() ) return;

    // get job
    Job local( model_.get( index ) );
    if( job.has( Job::SUB_TIME ) ) local.assign( Job::SUB_TIME, job.get( Job::SUB_TIME ) );
    if( job.has( Job::RUN_TIME ) ) local.assign( Job::RUN_TIME, job.get( Job::RUN_TIME ) );
    if( job.has( Job::END_TIME ) ) local.assign( Job::END_TIME, job.get( Job::END_TIME ) );

    if( job.hasParent() && local.parentId() != job.parentId() )
    {


        local.setParentId( job.parentId() );

        // need to re-run "checkChildren"
        JobModel::List currentJobs( model_.children() );
        JobSet jobs( JobSet::fromList( currentJobs ) );
        jobs.remove( local );
        jobs.insert( local );
        jobs.checkChildren();

        // re-emit
        emit jobListAvailable( user(), jobs );

        // assign to model
        model_.set( jobs.toList() );

    } else {

        // directly insert in list
        Debug::Throw( "JobManager::updateJob - inserting local.\n" );
        model_.add( local );

    }

}

//________________________________________
JobSet JobManager::selectedJobs( bool check )
{
    Debug::Throw( "JobManager::selectedJobs.\n" );
    JobSet selected;

    // retrieve selected indexes in list
    QModelIndexList selectedIndexes( list().selectionModel()->selectedIndexes() );
    for( QModelIndexList::iterator iter = selectedIndexes.begin(); iter != selectedIndexes.end(); ++iter )
    {

        // check column
        if( !iter->column() == 0 ) continue;
        if( !list().isVisible( *iter ) ) continue;

        // get current job
        const Job& job( model_.get( *iter ) );

        selected.insert( job );

    }

    if( check && selected.empty() )
    {
        // popup warning when no job is selected (nor removed by "permission denied" )
        QString buffer( "No selected jobs." );
        InformationDialog( this, buffer ).exec();
    }

    return selected;

}

//________________________________________
JobSet JobManager::masterJobs( void )
{
    Debug::Throw( "JobManager::masterJobs.\n" );
    JobSet masterJobs;
    foreach( const Job& job, model_.children() )
    { if( job.isMasterJob() ) masterJobs.insert( job ); }

    return masterJobs;

}

//____________________________________________________
void JobManager::_updateTimeStamps( JobSet& jobs, bool emitProgress )
{

    Debug::Throw( "JobManager::_updateTimeStamps.\n" );

    // retrieve current set
    JobModel::List currentJobs( model_.children() );

    if( currentJobs.empty() ) return;
    if( emitProgress ) emit busy( currentJobs.size() );
    int index(0);
    foreach( const Job& currentJob, currentJobs )
    {

        // progress
        if( emitProgress ) emit progressAvailable( index );
        ++index;

        // skip if timestamps are not set
        if( !( currentJob.has( Job::SUB_TIME ) || currentJob.has( Job::RUN_TIME ) || currentJob.has( Job::END_TIME ) || currentJob.hasParent() ) )
        { continue; }

        // find in current job set
        JobSet::const_iterator iter( jobs.find( currentJob ) );
        if( iter != jobs.end() )
        {
            Job job( *iter );
            bool changed( false );

            // copy parent if needed
            if( currentJob.hasParent() && currentJob.parentId() != iter->parentId() )
            { changed = true; job.setParentId( currentJob.parentId() ); }

            // copy timestamps if needed
            if( currentJob.has( Job::SUB_TIME ) && !iter->has( Job::SUB_TIME ) ) { changed = true; job.assign( Job::SUB_TIME, currentJob.get( Job::SUB_TIME ) ); }
            if( currentJob.has( Job::RUN_TIME ) && !iter->has( Job::RUN_TIME ) ) { changed = true; job.assign( Job::RUN_TIME, currentJob.get( Job::RUN_TIME ) ); }
            if( currentJob.has( Job::END_TIME ) && !iter->has( Job::END_TIME ) ) { changed = true; job.assign( Job::END_TIME, currentJob.get( Job::END_TIME ) ); }

            if( changed )
            {
                jobs.remove( *iter );
                jobs.insert( job );
            }

        }

    }

    if( emitProgress ) emit idle();

}


//____________________________________________________
void JobManager::_updateChildrenStatistics( JobSet& jobs )
{

    Debug::Throw( "JobManager::_updateChildrenStatistics.\n" );

    // retrieve current set
    foreach( const Job& currentJob, model_.children() )
    {

        // skip if timestamps are not set
        if( !currentJob.isMasterJob() )
        { continue; }

        // find in current job set
        JobSet::const_iterator iter( jobs.find( currentJob ) );
        if( iter != jobs.end() )
        {
            Job job( *iter );
            job.setChildrenStatistics( currentJob.childrenStatistics() );
            jobs.remove( *iter );
            jobs.insert( job );

        }

    }

}

//____________________________________________________
void JobManager::_updateChildrenStatistics( Job master, JobSet jobs )
{

    Debug::Throw( "JobManager::_updateChildrenStatistics.\n" );

    // insert master in joblist, check children and update
    jobs.insert( master );
    jobs.checkChildren();
    master = *jobs.find( master );

    // retrive model job
    Job modelJob( model_.get( model_.index( master ) ) );
    modelJob.setChildrenStatistics( master.childrenStatistics() );
    model_.add( modelJob );

}

//_____________________________________________________________
JobSet JobManager::_incompleteJobs( const JobSet& jobs ) const
{

    JobSet out;
    foreach( const Job& job, jobs )
    {

        // skip killed jobs, since they will dissapear
        if( job.isKilled() ) continue;

        // check for missing submission time
        if( !job.has( Job::SUB_TIME ) ) out.insert( job );

        // check for missing run time
        if( ( ( job.isRunning() && !job.isInserting() ) || job.isCompleted() || job.isFailed() ) && !job.has( Job::RUN_TIME ) ) out.insert( job );

        // check for missing end time
        else if( ( job.isCompleted() || job.isFailed() ) && !job.has( Job::END_TIME ) ) out.insert( job );

    }

    return out;
}

//________________________________________
void JobManager::_storeSelectedJobs( void )
{

    // clear
    model_.clearSelectedIndexes();

    // retrieve selected indexes in list
    foreach( const QModelIndex& index, list().selectionModel()->selectedIndexes() )
    { if( index.column() == 0 ) model_.setIndexSelected( index, true ); }

}

//________________________________________
void JobManager::_restoreSelectedJobs( void )
{

    QModelIndexList selectedIndexes( model_.selectedIndexes() );
    if( selectedIndexes.empty() ) list().selectionModel()->clear();
    else {

        list().selectionModel()->select( selectedIndexes.front(),  QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
        foreach( const QModelIndex& index, selectedIndexes )
        { list().selectionModel()->select( index, QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

    }

    return;
}

//________________________________________
void JobManager::_storeExpandedJobs( void )
{

    // clear
    model_.clearExpandedIndexes();

    // retrieve selected indexes in list
    foreach( const QModelIndex& index, model_.indexes() )
    { if( list().isExpanded( index ) ) model_.setIndexExpanded( index, true ); }

    list().storeScrollBarPosition();
    _lockRequests();
}

//________________________________________
void JobManager::_restoreExpandedJobs( void )
{

    // collapse all
    list().collapseAll();

    // restore expanded indices
    foreach( const QModelIndex& index, model_.expandedIndexes() )
    { list().setExpanded( index, true ); }

    list().restoreScrollBarPosition();
    _unlockRequests();

}

//________________________________________
void JobManager::_requestChildrenList( const QModelIndex& index )
{

    // check lock
    if( requestsLocked_ ) return;
    Debug::Throw( "JobManager::_requestChildrenList.\n" );

    // check index
    if( !index.isValid() ) return;

    // retrieve job, check if master and request children list
    const Job job( model_.get( index ) );
    if( job.isMasterJob() ) emit childrenListRequested( job );

}
