#ifndef _MainWindow_h_
#define _MainWindow_h_

// $Id: MainWindow.h,v 1.35 2012/07/23 15:13:33 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "AnimatedTabWidget.h"
#include "BaseMainWindow.h"
#include "Counter.h"
#include "CustomToolBar.h"
#include "Debug.h"
#include "Job.h"
#include "AlienCommand.h"
#include "Key.h"
#include "UserSet.h"

#include <QtGui/QCloseEvent>

class JobManagerFrame;
class JobSet;
class JobStatusBar;
class SummaryFrame;
class UserSelectionFrame;

//! condorwatch specific namespace
namespace TOP
{

    class TabWidget: public AnimatedTabWidget
    {

        public:

        //! constructor
        TabWidget( QWidget* parent ):
            AnimatedTabWidget( parent )
        {}

        protected:

        //! tab insertion
        virtual void tabInserted( int );

        //! tab insertion
        virtual void tabRemoved( int );

    };

};

/*!
\class MainWindow
\brief customized QVBox to recieve close event
*/
class MainWindow: public BaseMainWindow, public BASE::Key
{

    //! Qt meta object declaration
    Q_OBJECT

    public:

    //! constructor
    MainWindow( QWidget* parent = 0 );

    //! destructor
    virtual ~MainWindow( void );

    //! user
    UserSelectionFrame& userSelectionFrame( void ) const
    {
        assert( userSelectionFrame_ );
        return *userSelectionFrame_;
    }

    //! toolbar
    CustomToolBar& userToolBar( void )
    {
        assert( userToolBar_ );
        return *userToolBar_;
    }

    //! toolbar
    CustomToolBar& toolBar( void )
    {
        assert( toolbar_ );
        return *toolbar_;
    }

    //! status bar
    JobStatusBar& statusBar( void ) const
    {
        assert( statusbar_ );
        return *statusbar_;
    }

    //! install frames
    void installFrames( void );

    //!@name actions
    //@{

    //! default user action
    QAction& defaultUserAction( void ) const
    { return *defaultUserAction_; }

    //! user list action
    QAction& userListAction( void ) const
    { return *userListAction_; }

    //! toggle summary
    QAction& toggleSummaryAction( void ) const
    { return *toggleSummaryAction_; }

    //! update
    QAction& updateAction( void ) const
    { return *updateAction_; }

    //! resubmit
    QAction& resubmitAction( void ) const
    { return *resubmitAction_; }

    //! retrace
    QAction& traceAction( void ) const
    { return *traceAction_; }

    //! kill
    QAction& killAction( void ) const
    { return *killAction_; }

    //! add frame
    QAction& addFrameAction( void ) const
    { return *addFrameAction_; }

    //! remove current frame
    QAction& removeFrameAction( void ) const
    { return *removeFrameAction_; }

    //! clear
    QAction& clearAction( void ) const
    { return *clearAction_; }

    //@}

    //! install toolbar visibility actions
    /*! returns true if lockable toolbars are found */
    virtual bool installToolBarsActions( QMenu& );

    signals:

    //! request command
    void commandRequested( const AlienCommand& );

    //! request commands
    void commandsRequested( const AlienCommand::List& );

    //! request clear queue
    void clearQueueRequested( const QString&, const Job::IdSet& );

    //! request job information
    void jobInformationRequested( const Job& );

    //! user list ready
    void userListAvailable( QSet<QString> );

    //! active users have changed
    void activeUsersChanged( UserSet );

    public slots:

    //! request default user
    void requestDefaultUser( void );

    //! request user list
    void requestUserList( void );

    //! request job list
    void requestJobList( void );

    //! store default user
    void updateDefaultUser( QString );

    //! process user list
    void processUserList( UserSet );

    //! process job list
    void processMasterList( QString, const JobSet& );

    //! process job list
    void processChildrenList( const Job&, const JobSet& );

    //! process job information
    void processJobInformation( const Job& );

    //! process completed command
    void commandCompleted( const AlienCommand& );

    protected slots:

    //! update configuration
    void _updateConfiguration( void );

    //! save configuration
    void _saveConfiguration( void );

    //! update actions visibility
    void _updateActions( void );

    //! update clear action
    void _updateClearAction( void );

    //! update uesr
    void _updateUser( QString );

    //! active frame changed
    void _activeFrameChanged( int );

    //! toggle summary
    void _toggleSummary( bool );

    //! add frame
    void _addFrame( void );

    //! remove frame
    void _removeFrame( void );

    //! resubmit jobs
    void _resubmit( void );

    //! request children list
    void _requestChildrenList( const Job& );

    //! retrace jobs
    void _requestJobInformation( void );

    //! kill jobs
    void _kill( void );

    //! clear completed jobs
    void _clear( void );

    protected:

    //! timer event
    void timerEvent( QTimerEvent* );

    //! close window event handler
    void closeEvent( QCloseEvent* );

    //! active frame
    bool _hasActiveFrame( void ) const
    { return (bool) activeFrame_; }

    //! active frame
    JobManagerFrame& _activeFrame( void ) const
    { return *activeFrame_; }

    //! true if frame is active
    bool _isActiveFrame( const JobManagerFrame& frame ) const
    { return &frame == activeFrame_; }

    //! find frame matching user
    /*! return 0 if none found */
    JobManagerFrame* _findFrame( const QString& user ) const;

    //! active frame
    bool _activeFrameIsDefault( void ) const;

    //! request job list
    void _requestJobList( QString );

    //! request clear queue
    void _requestClearQueue( QString );

    //! active frame
    void _setActiveFrame( JobManagerFrame& frame );

    //! create a new frame
    JobManagerFrame& _newFrame( void );

    //! tab widget
    TOP::TabWidget& _tabWidget( void ) const
    { return *tabWidget_; }

    //! timer
    QBasicTimer& _timer( void )
    { return timer_; }

    //! get list of all active users
    UserSet _users( void ) const;

    //! true if default user is set
    bool _hasDefaultUser( void ) const
    { return !defaultUser_.isEmpty(); }

    private:

    //! summary
    void _installActions( void );

    //! toolbar
    CustomToolBar* userToolBar_;

    //! toolbar
    CustomToolBar* toolbar_;

    //! user
    UserSelectionFrame* userSelectionFrame_;

    //! status bar
    JobStatusBar* statusbar_;

    //! tab widget
    TOP::TabWidget* tabWidget_;

    //! active frame
    JobManagerFrame* activeFrame_;

    //! refresh rate
    int refreshRate_;

    //! timer object for scheduled update
    QBasicTimer timer_;

    //! default user
    QString defaultUser_;

    //!@name actions
    //@{

    //! default user action
    QAction* defaultUserAction_;

    //! user list action
    QAction* userListAction_;

    //! update
    QAction* updateAction_;

    //! resubmit
    QAction* resubmitAction_;

    //! retrace
    QAction* traceAction_;

    //! kill action
    QAction* killAction_;

    //! toggle summary frame
    QAction* toggleSummaryAction_;

    //! add frame
    QAction* addFrameAction_;

    //! remove frame
    QAction* removeFrameAction_;

    //! clear
    QAction* clearAction_;

    //@}

};

#endif
